# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional
from datetime import datetime

from .dispute import Dispute
from .._models import BaseModel
from .currency import Currency
from .country_code import CountryCode
from .intent_status import IntentStatus
from .billing_address import BillingAddress
from .refund_list_item import RefundListItem
from .custom_field_response import CustomFieldResponse
from .payment_refund_status import PaymentRefundStatus
from .customer_limited_details import CustomerLimitedDetails
from .one_time_product_cart_item import OneTimeProductCartItem

__all__ = ["Payment"]


class Payment(BaseModel):
    billing: BillingAddress
    """Billing address details for payments"""

    brand_id: str
    """brand id this payment belongs to"""

    business_id: str
    """Identifier of the business associated with the payment"""

    created_at: datetime
    """Timestamp when the payment was created"""

    currency: Currency
    """Currency used for the payment"""

    customer: CustomerLimitedDetails
    """Details about the customer who made the payment"""

    digital_products_delivered: bool
    """brand id this payment belongs to"""

    disputes: List[Dispute]
    """List of disputes associated with this payment"""

    metadata: Dict[str, str]
    """Additional custom data associated with the payment"""

    payment_id: str
    """Unique identifier for the payment"""

    refunds: List[RefundListItem]
    """List of refunds issued for this payment"""

    settlement_amount: int
    """
    The amount that will be credited to your Dodo balance after currency conversion
    and processing. Especially relevant for adaptive pricing where the customer's
    payment currency differs from your settlement currency.
    """

    settlement_currency: Currency
    """
    The currency in which the settlement_amount will be credited to your Dodo
    balance. This may differ from the customer's payment currency in adaptive
    pricing scenarios.
    """

    total_amount: int
    """
    Total amount charged to the customer including tax, in smallest currency unit
    (e.g. cents)
    """

    card_holder_name: Optional[str] = None
    """Cardholder name"""

    card_issuing_country: Optional[CountryCode] = None
    """ISO2 country code of the card"""

    card_last_four: Optional[str] = None
    """The last four digits of the card"""

    card_network: Optional[str] = None
    """Card network like VISA, MASTERCARD etc."""

    card_type: Optional[str] = None
    """The type of card DEBIT or CREDIT"""

    checkout_session_id: Optional[str] = None
    """
    If payment is made using a checkout session, this field is set to the id of the
    session.
    """

    custom_field_responses: Optional[List[CustomFieldResponse]] = None
    """Customer's responses to custom fields collected during checkout"""

    discount_id: Optional[str] = None
    """The discount id if discount is applied"""

    error_code: Optional[str] = None
    """An error code if the payment failed"""

    error_message: Optional[str] = None
    """An error message if the payment failed"""

    invoice_id: Optional[str] = None
    """Invoice ID for this payment. Uses India-specific invoice ID if available."""

    invoice_url: Optional[str] = None
    """URL to download the invoice PDF for this payment."""

    payment_link: Optional[str] = None
    """Checkout URL"""

    payment_method: Optional[str] = None
    """Payment method used by customer (e.g. "card", "bank_transfer")"""

    payment_method_type: Optional[str] = None
    """Specific type of payment method (e.g. "visa", "mastercard")"""

    product_cart: Optional[List[OneTimeProductCartItem]] = None
    """List of products purchased in a one-time payment"""

    refund_status: Optional[PaymentRefundStatus] = None
    """Summary of the refund status for this payment.

    None if no succeeded refunds exist.
    """

    settlement_tax: Optional[int] = None
    """
    This represents the portion of settlement_amount that corresponds to taxes
    collected. Especially relevant for adaptive pricing where the tax component must
    be tracked separately in your Dodo balance.
    """

    status: Optional[IntentStatus] = None
    """Current status of the payment intent"""

    subscription_id: Optional[str] = None
    """Identifier of the subscription if payment is part of a subscription"""

    tax: Optional[int] = None
    """Amount of tax collected in smallest currency unit (e.g. cents)"""

    updated_at: Optional[datetime] = None
    """Timestamp when the payment was last updated"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .._models import BaseModel
from .time_interval import TimeInterval
from .cbb_overage_behavior import CbbOverageBehavior

__all__ = ["CreditEntitlementCartResponse"]


class CreditEntitlementCartResponse(BaseModel):
    """Response struct representing credit entitlement cart details for a subscription"""

    credit_entitlement_id: str

    credit_entitlement_name: str

    credits_amount: str

    overage_balance: str
    """Customer's current overage balance for this entitlement"""

    overage_behavior: CbbOverageBehavior
    """Controls how overage is handled at the end of a billing cycle.

    | Preset                     | Charge at billing | Credits reduce overage | Preserve overage at reset |
    | -------------------------- | :---------------: | :--------------------: | :-----------------------: |
    | `forgive_at_reset`         |        No         |           No           |            No             |
    | `invoice_at_billing`       |        Yes        |           No           |            No             |
    | `carry_deficit`            |        No         |           No           |            Yes            |
    | `carry_deficit_auto_repay` |        No         |          Yes           |            Yes            |
    """

    overage_enabled: bool

    product_id: str

    remaining_balance: str
    """Customer's current remaining credit balance for this entitlement"""

    rollover_enabled: bool

    unit: str
    """Unit label for the credit entitlement (e.g., "API Calls", "Tokens")"""

    expires_after_days: Optional[int] = None

    low_balance_threshold_percent: Optional[int] = None

    max_rollover_count: Optional[int] = None

    overage_limit: Optional[str] = None

    rollover_percentage: Optional[int] = None

    rollover_timeframe_count: Optional[int] = None

    rollover_timeframe_interval: Optional[TimeInterval] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime

from .._models import BaseModel
from .discount_type import DiscountType

__all__ = ["Discount"]


class Discount(BaseModel):
    amount: int
    """The discount amount.

    - If `discount_type` is `percentage`, this is in **basis points** (e.g., 540 =>
      5.4%).
    - Otherwise, this is **USD cents** (e.g., 100 => `$1.00`).
    """

    business_id: str
    """The business this discount belongs to."""

    code: str
    """The discount code (up to 16 chars)."""

    created_at: datetime
    """Timestamp when the discount is created"""

    discount_id: str
    """The unique discount ID"""

    preserve_on_plan_change: bool
    """
    Whether this discount should be preserved when a subscription changes plans.
    Default: false (discount is removed on plan change)
    """

    restricted_to: List[str]
    """List of product IDs to which this discount is restricted."""

    times_used: int
    """How many times this discount has been used."""

    type: DiscountType
    """The type of discount, e.g. `percentage`, `flat`, or `flat_per_unit`."""

    expires_at: Optional[datetime] = None
    """Optional date/time after which discount is expired."""

    name: Optional[str] = None
    """Name for the Discount"""

    subscription_cycles: Optional[int] = None
    """
    Number of subscription billing cycles this discount is valid for. If not
    provided, the discount will be applied indefinitely to all recurring payments
    related to the subscription.
    """

    usage_limit: Optional[int] = None
    """Usage limit for this discount, if any."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import TypedDict

from .on_demand_subscription_param import OnDemandSubscriptionParam

__all__ = ["SubscriptionDataParam"]


class SubscriptionDataParam(TypedDict, total=False):
    on_demand: Optional[OnDemandSubscriptionParam]

    trial_period_days: Optional[int]
    """
    Optional trial period in days If specified, this value overrides the trial
    period set in the product's price Must be between 0 and 10000 days
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["PaymentRefundStatus"]

PaymentRefundStatus: TypeAlias = Literal["partial", "full"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["MeterCreditEntitlementCartResponse"]


class MeterCreditEntitlementCartResponse(BaseModel):
    """
    Response struct representing meter-credit entitlement mapping cart details for a subscription
    """

    credit_entitlement_id: str

    meter_id: str

    meter_name: str

    meter_units_per_credit: str

    product_id: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

from .discount_type import DiscountType

__all__ = ["DiscountListParams"]


class DiscountListParams(TypedDict, total=False):
    active: bool
    """Filter by active status (true = not expired, false = expired)"""

    code: str
    """Filter by discount code (partial match, case-insensitive)"""

    discount_type: DiscountType
    """Filter by discount type (percentage)"""

    page_number: int
    """Page number (default = 0)."""

    page_size: int
    """Page size (default = 10, max = 100)."""

    product_id: str
    """Filter by product restriction (only discounts that apply to this product)"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional
from typing_extensions import TypedDict

__all__ = ["CustomerUpdateParams"]


class CustomerUpdateParams(TypedDict, total=False):
    email: Optional[str]

    metadata: Optional[Dict[str, str]]
    """Additional metadata for the customer"""

    name: Optional[str]

    phone_number: Optional[str]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .._models import BaseModel

__all__ = ["CheckoutSessionResponse"]


class CheckoutSessionResponse(BaseModel):
    session_id: str
    """The ID of the created checkout session"""

    checkout_url: Optional[str] = None
    """Checkout url (None when payment_method_id is provided)"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = ["MeterAggregationParam"]


class MeterAggregationParam(TypedDict, total=False):
    type: Required[Literal["count", "sum", "max", "last"]]
    """Aggregation type for the meter"""

    key: Optional[str]
    """Required when type is not COUNT"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from .._models import BaseModel
from .currency import Currency
from .time_interval import TimeInterval
from .cbb_overage_behavior import CbbOverageBehavior

__all__ = ["CreditEntitlement"]


class CreditEntitlement(BaseModel):
    id: str

    business_id: str

    created_at: datetime

    name: str

    overage_behavior: CbbOverageBehavior
    """Controls how overage is handled at billing cycle end."""

    overage_enabled: bool

    precision: int

    rollover_enabled: bool

    unit: str

    updated_at: datetime

    currency: Optional[Currency] = None

    description: Optional[str] = None

    expires_after_days: Optional[int] = None

    max_rollover_count: Optional[int] = None

    overage_limit: Optional[int] = None

    price_per_unit: Optional[str] = None
    """Price per credit unit"""

    rollover_percentage: Optional[int] = None

    rollover_timeframe_count: Optional[int] = None

    rollover_timeframe_interval: Optional[TimeInterval] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import httpx
import pytest
from respx import MockRouter

from dodopayments import DodoPayments, AsyncDodoPayments
from dodopayments._response import (
    BinaryAPIResponse,
    AsyncBinaryAPIResponse,
    StreamedBinaryAPIResponse,
    AsyncStreamedBinaryAPIResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestPayments:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_method_retrieve(self, client: DodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/payments/payment_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        payment = client.invoices.payments.retrieve(
            "payment_id",
        )
        assert payment.is_closed
        assert payment.json() == {"foo": "bar"}
        assert cast(Any, payment.is_closed) is True
        assert isinstance(payment, BinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_raw_response_retrieve(self, client: DodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/payments/payment_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))

        payment = client.invoices.payments.with_raw_response.retrieve(
            "payment_id",
        )

        assert payment.is_closed is True
        assert payment.http_request.headers.get("X-Stainless-Lang") == "python"
        assert payment.json() == {"foo": "bar"}
        assert isinstance(payment, BinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_streaming_response_retrieve(self, client: DodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/payments/payment_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        with client.invoices.payments.with_streaming_response.retrieve(
            "payment_id",
        ) as payment:
            assert not payment.is_closed
            assert payment.http_request.headers.get("X-Stainless-Lang") == "python"

            assert payment.json() == {"foo": "bar"}
            assert cast(Any, payment.is_closed) is True
            assert isinstance(payment, StreamedBinaryAPIResponse)

        assert cast(Any, payment.is_closed) is True

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_path_params_retrieve(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `payment_id` but received ''"):
            client.invoices.payments.with_raw_response.retrieve(
                "",
            )

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_method_retrieve_refund(self, client: DodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/refunds/refund_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        payment = client.invoices.payments.retrieve_refund(
            "refund_id",
        )
        assert payment.is_closed
        assert payment.json() == {"foo": "bar"}
        assert cast(Any, payment.is_closed) is True
        assert isinstance(payment, BinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_raw_response_retrieve_refund(self, client: DodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/refunds/refund_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))

        payment = client.invoices.payments.with_raw_response.retrieve_refund(
            "refund_id",
        )

        assert payment.is_closed is True
        assert payment.http_request.headers.get("X-Stainless-Lang") == "python"
        assert payment.json() == {"foo": "bar"}
        assert isinstance(payment, BinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_streaming_response_retrieve_refund(self, client: DodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/refunds/refund_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        with client.invoices.payments.with_streaming_response.retrieve_refund(
            "refund_id",
        ) as payment:
            assert not payment.is_closed
            assert payment.http_request.headers.get("X-Stainless-Lang") == "python"

            assert payment.json() == {"foo": "bar"}
            assert cast(Any, payment.is_closed) is True
            assert isinstance(payment, StreamedBinaryAPIResponse)

        assert cast(Any, payment.is_closed) is True

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_path_params_retrieve_refund(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `refund_id` but received ''"):
            client.invoices.payments.with_raw_response.retrieve_refund(
                "",
            )


class TestAsyncPayments:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_method_retrieve(self, async_client: AsyncDodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/payments/payment_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        payment = await async_client.invoices.payments.retrieve(
            "payment_id",
        )
        assert payment.is_closed
        assert await payment.json() == {"foo": "bar"}
        assert cast(Any, payment.is_closed) is True
        assert isinstance(payment, AsyncBinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_raw_response_retrieve(self, async_client: AsyncDodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/payments/payment_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))

        payment = await async_client.invoices.payments.with_raw_response.retrieve(
            "payment_id",
        )

        assert payment.is_closed is True
        assert payment.http_request.headers.get("X-Stainless-Lang") == "python"
        assert await payment.json() == {"foo": "bar"}
        assert isinstance(payment, AsyncBinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_streaming_response_retrieve(self, async_client: AsyncDodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/payments/payment_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        async with async_client.invoices.payments.with_streaming_response.retrieve(
            "payment_id",
        ) as payment:
            assert not payment.is_closed
            assert payment.http_request.headers.get("X-Stainless-Lang") == "python"

            assert await payment.json() == {"foo": "bar"}
            assert cast(Any, payment.is_closed) is True
            assert isinstance(payment, AsyncStreamedBinaryAPIResponse)

        assert cast(Any, payment.is_closed) is True

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_path_params_retrieve(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `payment_id` but received ''"):
            await async_client.invoices.payments.with_raw_response.retrieve(
                "",
            )

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_method_retrieve_refund(self, async_client: AsyncDodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/refunds/refund_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        payment = await async_client.invoices.payments.retrieve_refund(
            "refund_id",
        )
        assert payment.is_closed
        assert await payment.json() == {"foo": "bar"}
        assert cast(Any, payment.is_closed) is True
        assert isinstance(payment, AsyncBinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_raw_response_retrieve_refund(self, async_client: AsyncDodoPayments, respx_mock: MockRouter) -> None:
        respx_mock.get("/invoices/refunds/refund_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))

        payment = await async_client.invoices.payments.with_raw_response.retrieve_refund(
            "refund_id",
        )

        assert payment.is_closed is True
        assert payment.http_request.headers.get("X-Stainless-Lang") == "python"
        assert await payment.json() == {"foo": "bar"}
        assert isinstance(payment, AsyncBinaryAPIResponse)

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_streaming_response_retrieve_refund(
        self, async_client: AsyncDodoPayments, respx_mock: MockRouter
    ) -> None:
        respx_mock.get("/invoices/refunds/refund_id").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        async with async_client.invoices.payments.with_streaming_response.retrieve_refund(
            "refund_id",
        ) as payment:
            assert not payment.is_closed
            assert payment.http_request.headers.get("X-Stainless-Lang") == "python"

            assert await payment.json() == {"foo": "bar"}
            assert cast(Any, payment.is_closed) is True
            assert isinstance(payment, AsyncStreamedBinaryAPIResponse)

        assert cast(Any, payment.is_closed) is True

    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_path_params_retrieve_refund(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `refund_id` but received ''"):
            await async_client.invoices.payments.with_raw_response.retrieve_refund(
                "",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Optional
from typing_extensions import Required, TypedDict

from .webhook_event_type import WebhookEventType

__all__ = ["WebhookCreateParams"]


class WebhookCreateParams(TypedDict, total=False):
    url: Required[str]
    """Url of the webhook"""

    description: Optional[str]

    disabled: Optional[bool]
    """Create the webhook in a disabled state.

    Default is false
    """

    filter_types: List[WebhookEventType]
    """Filter events to the webhook.

    Webhook event will only be sent for events in the list.
    """

    headers: Optional[Dict[str, str]]
    """Custom headers to be passed"""

    idempotency_key: Optional[str]
    """The request's idempotency key"""

    metadata: Optional[Dict[str, str]]
    """Metadata to be passed to the webhook Defaut is {}"""

    rate_limit: Optional[int]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal, Required, TypedDict

from .._types import SequenceNotStr

__all__ = ["CustomFieldParam"]


class CustomFieldParam(TypedDict, total=False):
    """Definition of a custom field for checkout"""

    field_type: Required[Literal["text", "number", "email", "url", "date", "dropdown", "boolean"]]
    """Type of field determining validation rules"""

    key: Required[str]
    """Unique identifier for this field (used as key in responses)"""

    label: Required[str]
    """Display label shown to customer"""

    options: Optional[SequenceNotStr[str]]
    """Options for dropdown type (required for dropdown, ignored for others)"""

    placeholder: Optional[str]
    """Placeholder text for the input"""

    required: bool
    """Whether this field is required"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from tests.utils import assert_matches_type
from dodopayments import DodoPayments, AsyncDodoPayments
from dodopayments.types import (
    Subscription,
    SubscriptionListResponse,
    SubscriptionChargeResponse,
    SubscriptionCreateResponse,
    SubscriptionPreviewChangePlanResponse,
    SubscriptionRetrieveCreditUsageResponse,
    SubscriptionUpdatePaymentMethodResponse,
    SubscriptionRetrieveUsageHistoryResponse,
)
from dodopayments._utils import parse_datetime
from dodopayments.pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination

# pyright: reportDeprecated=false

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestSubscriptions:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_create(self, client: DodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            subscription = client.subscriptions.create(
                billing={"country": "AF"},
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
            )

        assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

    @parametrize
    def test_method_create_with_all_params(self, client: DodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            subscription = client.subscriptions.create(
                billing={
                    "country": "AF",
                    "city": "city",
                    "state": "state",
                    "street": "street",
                    "zipcode": "zipcode",
                },
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
                addons=[
                    {
                        "addon_id": "addon_id",
                        "quantity": 0,
                    }
                ],
                allowed_payment_method_types=["ach"],
                billing_currency="AED",
                discount_code="discount_code",
                force_3ds=True,
                metadata={"foo": "string"},
                on_demand={
                    "mandate_only": True,
                    "adaptive_currency_fees_inclusive": True,
                    "product_currency": "AED",
                    "product_description": "product_description",
                    "product_price": 0,
                },
                one_time_product_cart=[
                    {
                        "product_id": "product_id",
                        "quantity": 0,
                        "amount": 0,
                    }
                ],
                payment_link=True,
                payment_method_id="payment_method_id",
                redirect_immediately=True,
                return_url="return_url",
                short_link=True,
                show_saved_payment_methods=True,
                tax_id="tax_id",
                trial_period_days=0,
            )

        assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

    @parametrize
    def test_raw_response_create(self, client: DodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            response = client.subscriptions.with_raw_response.create(
                billing={"country": "AF"},
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
            )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

    @parametrize
    def test_streaming_response_create(self, client: DodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            with client.subscriptions.with_streaming_response.create(
                billing={"country": "AF"},
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
            ) as response:
                assert not response.is_closed
                assert response.http_request.headers.get("X-Stainless-Lang") == "python"

                subscription = response.parse()
                assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_retrieve(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.retrieve(
            "subscription_id",
        )
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    def test_raw_response_retrieve(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.retrieve(
            "subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    def test_streaming_response_retrieve(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.retrieve(
            "subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(Subscription, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.retrieve(
                "",
            )

    @parametrize
    def test_method_update(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.update(
            subscription_id="subscription_id",
        )
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    def test_method_update_with_all_params(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.update(
            subscription_id="subscription_id",
            billing={
                "country": "AF",
                "city": "city",
                "state": "state",
                "street": "street",
                "zipcode": "zipcode",
            },
            cancel_at_next_billing_date=True,
            credit_entitlement_cart=[
                {
                    "credit_entitlement_id": "credit_entitlement_id",
                    "credits_amount": "credits_amount",
                    "expires_after_days": 0,
                    "low_balance_threshold_percent": 0,
                    "max_rollover_count": 0,
                    "overage_enabled": True,
                    "overage_limit": "overage_limit",
                    "rollover_enabled": True,
                    "rollover_percentage": 0,
                    "rollover_timeframe_count": 0,
                    "rollover_timeframe_interval": "Day",
                }
            ],
            customer_name="customer_name",
            disable_on_demand={"next_billing_date": parse_datetime("2019-12-27T18:11:19.117Z")},
            metadata={"foo": "string"},
            next_billing_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            status="pending",
            tax_id="tax_id",
        )
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    def test_raw_response_update(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.update(
            subscription_id="subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    def test_streaming_response_update(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.update(
            subscription_id="subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(Subscription, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.update(
                subscription_id="",
            )

    @parametrize
    def test_method_list(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.list()
        assert_matches_type(SyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.list(
            brand_id="brand_id",
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            customer_id="customer_id",
            page_number=0,
            page_size=0,
            product_id="product_id",
            status="pending",
        )
        assert_matches_type(SyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(SyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(
                SyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"]
            )

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_change_plan(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )
        assert subscription is None

    @parametrize
    def test_method_change_plan_with_all_params(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
            addons=[
                {
                    "addon_id": "addon_id",
                    "quantity": 0,
                }
            ],
            discount_code="discount_code",
            metadata={"foo": "string"},
            on_payment_failure="prevent_change",
        )
        assert subscription is None

    @parametrize
    def test_raw_response_change_plan(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert subscription is None

    @parametrize
    def test_streaming_response_change_plan(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert subscription is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_change_plan(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.change_plan(
                subscription_id="",
                product_id="product_id",
                proration_billing_mode="prorated_immediately",
                quantity=0,
            )

    @parametrize
    def test_method_charge(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.charge(
            subscription_id="subscription_id",
            product_price=0,
        )
        assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

    @parametrize
    def test_method_charge_with_all_params(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.charge(
            subscription_id="subscription_id",
            product_price=0,
            adaptive_currency_fees_inclusive=True,
            customer_balance_config={
                "allow_customer_credits_purchase": True,
                "allow_customer_credits_usage": True,
            },
            metadata={"foo": "string"},
            product_currency="AED",
            product_description="product_description",
        )
        assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

    @parametrize
    def test_raw_response_charge(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.charge(
            subscription_id="subscription_id",
            product_price=0,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

    @parametrize
    def test_streaming_response_charge(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.charge(
            subscription_id="subscription_id",
            product_price=0,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_charge(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.charge(
                subscription_id="",
                product_price=0,
            )

    @parametrize
    def test_method_preview_change_plan(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )
        assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

    @parametrize
    def test_method_preview_change_plan_with_all_params(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
            addons=[
                {
                    "addon_id": "addon_id",
                    "quantity": 0,
                }
            ],
            discount_code="discount_code",
            metadata={"foo": "string"},
            on_payment_failure="prevent_change",
        )
        assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

    @parametrize
    def test_raw_response_preview_change_plan(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

    @parametrize
    def test_streaming_response_preview_change_plan(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_preview_change_plan(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.preview_change_plan(
                subscription_id="",
                product_id="product_id",
                proration_billing_mode="prorated_immediately",
                quantity=0,
            )

    @parametrize
    def test_method_retrieve_credit_usage(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.retrieve_credit_usage(
            "subscription_id",
        )
        assert_matches_type(SubscriptionRetrieveCreditUsageResponse, subscription, path=["response"])

    @parametrize
    def test_raw_response_retrieve_credit_usage(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.retrieve_credit_usage(
            "subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(SubscriptionRetrieveCreditUsageResponse, subscription, path=["response"])

    @parametrize
    def test_streaming_response_retrieve_credit_usage(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.retrieve_credit_usage(
            "subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(SubscriptionRetrieveCreditUsageResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve_credit_usage(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.retrieve_credit_usage(
                "",
            )

    @parametrize
    def test_method_retrieve_usage_history(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.retrieve_usage_history(
            subscription_id="subscription_id",
        )
        assert_matches_type(
            SyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse], subscription, path=["response"]
        )

    @parametrize
    def test_method_retrieve_usage_history_with_all_params(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.retrieve_usage_history(
            subscription_id="subscription_id",
            end_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            meter_id="meter_id",
            page_number=0,
            page_size=0,
            start_date=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(
            SyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse], subscription, path=["response"]
        )

    @parametrize
    def test_raw_response_retrieve_usage_history(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.retrieve_usage_history(
            subscription_id="subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(
            SyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse], subscription, path=["response"]
        )

    @parametrize
    def test_streaming_response_retrieve_usage_history(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.retrieve_usage_history(
            subscription_id="subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(
                SyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse],
                subscription,
                path=["response"],
            )

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve_usage_history(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.retrieve_usage_history(
                subscription_id="",
            )

    @parametrize
    def test_method_update_payment_method_overload_1(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.update_payment_method(
            subscription_id="subscription_id",
            type="new",
        )
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    def test_method_update_payment_method_with_all_params_overload_1(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.update_payment_method(
            subscription_id="subscription_id",
            type="new",
            return_url="return_url",
        )
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    def test_raw_response_update_payment_method_overload_1(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.update_payment_method(
            subscription_id="subscription_id",
            type="new",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    def test_streaming_response_update_payment_method_overload_1(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.update_payment_method(
            subscription_id="subscription_id",
            type="new",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update_payment_method_overload_1(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.update_payment_method(
                subscription_id="",
                type="new",
            )

    @parametrize
    def test_method_update_payment_method_overload_2(self, client: DodoPayments) -> None:
        subscription = client.subscriptions.update_payment_method(
            subscription_id="subscription_id",
            payment_method_id="payment_method_id",
            type="existing",
        )
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    def test_raw_response_update_payment_method_overload_2(self, client: DodoPayments) -> None:
        response = client.subscriptions.with_raw_response.update_payment_method(
            subscription_id="subscription_id",
            payment_method_id="payment_method_id",
            type="existing",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = response.parse()
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    def test_streaming_response_update_payment_method_overload_2(self, client: DodoPayments) -> None:
        with client.subscriptions.with_streaming_response.update_payment_method(
            subscription_id="subscription_id",
            payment_method_id="payment_method_id",
            type="existing",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = response.parse()
            assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update_payment_method_overload_2(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            client.subscriptions.with_raw_response.update_payment_method(
                subscription_id="",
                payment_method_id="payment_method_id",
                type="existing",
            )


class TestAsyncSubscriptions:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_create(self, async_client: AsyncDodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            subscription = await async_client.subscriptions.create(
                billing={"country": "AF"},
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
            )

        assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

    @parametrize
    async def test_method_create_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            subscription = await async_client.subscriptions.create(
                billing={
                    "country": "AF",
                    "city": "city",
                    "state": "state",
                    "street": "street",
                    "zipcode": "zipcode",
                },
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
                addons=[
                    {
                        "addon_id": "addon_id",
                        "quantity": 0,
                    }
                ],
                allowed_payment_method_types=["ach"],
                billing_currency="AED",
                discount_code="discount_code",
                force_3ds=True,
                metadata={"foo": "string"},
                on_demand={
                    "mandate_only": True,
                    "adaptive_currency_fees_inclusive": True,
                    "product_currency": "AED",
                    "product_description": "product_description",
                    "product_price": 0,
                },
                one_time_product_cart=[
                    {
                        "product_id": "product_id",
                        "quantity": 0,
                        "amount": 0,
                    }
                ],
                payment_link=True,
                payment_method_id="payment_method_id",
                redirect_immediately=True,
                return_url="return_url",
                short_link=True,
                show_saved_payment_methods=True,
                tax_id="tax_id",
                trial_period_days=0,
            )

        assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

    @parametrize
    async def test_raw_response_create(self, async_client: AsyncDodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            response = await async_client.subscriptions.with_raw_response.create(
                billing={"country": "AF"},
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
            )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_create(self, async_client: AsyncDodoPayments) -> None:
        with pytest.warns(DeprecationWarning):
            async with async_client.subscriptions.with_streaming_response.create(
                billing={"country": "AF"},
                customer={"customer_id": "customer_id"},
                product_id="product_id",
                quantity=0,
            ) as response:
                assert not response.is_closed
                assert response.http_request.headers.get("X-Stainless-Lang") == "python"

                subscription = await response.parse()
                assert_matches_type(SubscriptionCreateResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_retrieve(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.retrieve(
            "subscription_id",
        )
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.retrieve(
            "subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.retrieve(
            "subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(Subscription, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.retrieve(
                "",
            )

    @parametrize
    async def test_method_update(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.update(
            subscription_id="subscription_id",
        )
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    async def test_method_update_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.update(
            subscription_id="subscription_id",
            billing={
                "country": "AF",
                "city": "city",
                "state": "state",
                "street": "street",
                "zipcode": "zipcode",
            },
            cancel_at_next_billing_date=True,
            credit_entitlement_cart=[
                {
                    "credit_entitlement_id": "credit_entitlement_id",
                    "credits_amount": "credits_amount",
                    "expires_after_days": 0,
                    "low_balance_threshold_percent": 0,
                    "max_rollover_count": 0,
                    "overage_enabled": True,
                    "overage_limit": "overage_limit",
                    "rollover_enabled": True,
                    "rollover_percentage": 0,
                    "rollover_timeframe_count": 0,
                    "rollover_timeframe_interval": "Day",
                }
            ],
            customer_name="customer_name",
            disable_on_demand={"next_billing_date": parse_datetime("2019-12-27T18:11:19.117Z")},
            metadata={"foo": "string"},
            next_billing_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            status="pending",
            tax_id="tax_id",
        )
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    async def test_raw_response_update(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.update(
            subscription_id="subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(Subscription, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.update(
            subscription_id="subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(Subscription, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.update(
                subscription_id="",
            )

    @parametrize
    async def test_method_list(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.list()
        assert_matches_type(AsyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.list(
            brand_id="brand_id",
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            customer_id="customer_id",
            page_number=0,
            page_size=0,
            product_id="product_id",
            status="pending",
        )
        assert_matches_type(AsyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(AsyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(
                AsyncDefaultPageNumberPagination[SubscriptionListResponse], subscription, path=["response"]
            )

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_change_plan(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )
        assert subscription is None

    @parametrize
    async def test_method_change_plan_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
            addons=[
                {
                    "addon_id": "addon_id",
                    "quantity": 0,
                }
            ],
            discount_code="discount_code",
            metadata={"foo": "string"},
            on_payment_failure="prevent_change",
        )
        assert subscription is None

    @parametrize
    async def test_raw_response_change_plan(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert subscription is None

    @parametrize
    async def test_streaming_response_change_plan(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert subscription is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_change_plan(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.change_plan(
                subscription_id="",
                product_id="product_id",
                proration_billing_mode="prorated_immediately",
                quantity=0,
            )

    @parametrize
    async def test_method_charge(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.charge(
            subscription_id="subscription_id",
            product_price=0,
        )
        assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

    @parametrize
    async def test_method_charge_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.charge(
            subscription_id="subscription_id",
            product_price=0,
            adaptive_currency_fees_inclusive=True,
            customer_balance_config={
                "allow_customer_credits_purchase": True,
                "allow_customer_credits_usage": True,
            },
            metadata={"foo": "string"},
            product_currency="AED",
            product_description="product_description",
        )
        assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

    @parametrize
    async def test_raw_response_charge(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.charge(
            subscription_id="subscription_id",
            product_price=0,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_charge(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.charge(
            subscription_id="subscription_id",
            product_price=0,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(SubscriptionChargeResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_charge(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.charge(
                subscription_id="",
                product_price=0,
            )

    @parametrize
    async def test_method_preview_change_plan(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )
        assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

    @parametrize
    async def test_method_preview_change_plan_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
            addons=[
                {
                    "addon_id": "addon_id",
                    "quantity": 0,
                }
            ],
            discount_code="discount_code",
            metadata={"foo": "string"},
            on_payment_failure="prevent_change",
        )
        assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

    @parametrize
    async def test_raw_response_preview_change_plan(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_preview_change_plan(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.preview_change_plan(
            subscription_id="subscription_id",
            product_id="product_id",
            proration_billing_mode="prorated_immediately",
            quantity=0,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(SubscriptionPreviewChangePlanResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_preview_change_plan(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.preview_change_plan(
                subscription_id="",
                product_id="product_id",
                proration_billing_mode="prorated_immediately",
                quantity=0,
            )

    @parametrize
    async def test_method_retrieve_credit_usage(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.retrieve_credit_usage(
            "subscription_id",
        )
        assert_matches_type(SubscriptionRetrieveCreditUsageResponse, subscription, path=["response"])

    @parametrize
    async def test_raw_response_retrieve_credit_usage(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.retrieve_credit_usage(
            "subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(SubscriptionRetrieveCreditUsageResponse, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve_credit_usage(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.retrieve_credit_usage(
            "subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(SubscriptionRetrieveCreditUsageResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve_credit_usage(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.retrieve_credit_usage(
                "",
            )

    @parametrize
    async def test_method_retrieve_usage_history(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.retrieve_usage_history(
            subscription_id="subscription_id",
        )
        assert_matches_type(
            AsyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse], subscription, path=["response"]
        )

    @parametrize
    async def test_method_retrieve_usage_history_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.retrieve_usage_history(
            subscription_id="subscription_id",
            end_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            meter_id="meter_id",
            page_number=0,
            page_size=0,
            start_date=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(
            AsyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse], subscription, path=["response"]
        )

    @parametrize
    async def test_raw_response_retrieve_usage_history(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.retrieve_usage_history(
            subscription_id="subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(
            AsyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse], subscription, path=["response"]
        )

    @parametrize
    async def test_streaming_response_retrieve_usage_history(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.retrieve_usage_history(
            subscription_id="subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(
                AsyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse],
                subscription,
                path=["response"],
            )

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve_usage_history(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.retrieve_usage_history(
                subscription_id="",
            )

    @parametrize
    async def test_method_update_payment_method_overload_1(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.update_payment_method(
            subscription_id="subscription_id",
            type="new",
        )
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    async def test_method_update_payment_method_with_all_params_overload_1(
        self, async_client: AsyncDodoPayments
    ) -> None:
        subscription = await async_client.subscriptions.update_payment_method(
            subscription_id="subscription_id",
            type="new",
            return_url="return_url",
        )
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    async def test_raw_response_update_payment_method_overload_1(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.update_payment_method(
            subscription_id="subscription_id",
            type="new",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_update_payment_method_overload_1(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.update_payment_method(
            subscription_id="subscription_id",
            type="new",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update_payment_method_overload_1(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.update_payment_method(
                subscription_id="",
                type="new",
            )

    @parametrize
    async def test_method_update_payment_method_overload_2(self, async_client: AsyncDodoPayments) -> None:
        subscription = await async_client.subscriptions.update_payment_method(
            subscription_id="subscription_id",
            payment_method_id="payment_method_id",
            type="existing",
        )
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    async def test_raw_response_update_payment_method_overload_2(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.subscriptions.with_raw_response.update_payment_method(
            subscription_id="subscription_id",
            payment_method_id="payment_method_id",
            type="existing",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription = await response.parse()
        assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

    @parametrize
    async def test_streaming_response_update_payment_method_overload_2(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.subscriptions.with_streaming_response.update_payment_method(
            subscription_id="subscription_id",
            payment_method_id="payment_method_id",
            type="existing",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription = await response.parse()
            assert_matches_type(SubscriptionUpdatePaymentMethodResponse, subscription, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update_payment_method_overload_2(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `subscription_id` but received ''"):
            await async_client.subscriptions.with_raw_response.update_payment_method(
                subscription_id="",
                payment_method_id="payment_method_id",
                type="existing",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["CreditBalanceLowWebhookEvent", "Data"]


class Data(BaseModel):
    """Webhook payload for credit.balance_low event"""

    available_balance: str

    credit_entitlement_id: str

    credit_entitlement_name: str

    customer_id: str

    subscription_credits_amount: str

    subscription_id: str

    threshold_amount: str

    threshold_percent: int


class CreditBalanceLowWebhookEvent(BaseModel):
    business_id: str
    """The business identifier"""

    data: Data
    """Webhook payload for credit.balance_low event"""

    timestamp: datetime
    """The timestamp of when the event occurred"""

    type: Literal["credit.balance_low"]
    """The event type"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from tests.utils import assert_matches_type
from dodopayments import DodoPayments, AsyncDodoPayments
from dodopayments.types import BalanceLedgerEntry
from dodopayments._utils import parse_datetime
from dodopayments.pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestBalances:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_retrieve_ledger(self, client: DodoPayments) -> None:
        balance = client.balances.retrieve_ledger()
        assert_matches_type(SyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

    @parametrize
    def test_method_retrieve_ledger_with_all_params(self, client: DodoPayments) -> None:
        balance = client.balances.retrieve_ledger(
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            currency="AED",
            event_type="payment",
            limit=0,
            page_number=0,
            page_size=0,
            reference_object_id="reference_object_id",
        )
        assert_matches_type(SyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

    @parametrize
    def test_raw_response_retrieve_ledger(self, client: DodoPayments) -> None:
        response = client.balances.with_raw_response.retrieve_ledger()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        balance = response.parse()
        assert_matches_type(SyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

    @parametrize
    def test_streaming_response_retrieve_ledger(self, client: DodoPayments) -> None:
        with client.balances.with_streaming_response.retrieve_ledger() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            balance = response.parse()
            assert_matches_type(SyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncBalances:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_retrieve_ledger(self, async_client: AsyncDodoPayments) -> None:
        balance = await async_client.balances.retrieve_ledger()
        assert_matches_type(AsyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

    @parametrize
    async def test_method_retrieve_ledger_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        balance = await async_client.balances.retrieve_ledger(
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            currency="AED",
            event_type="payment",
            limit=0,
            page_number=0,
            page_size=0,
            reference_object_id="reference_object_id",
        )
        assert_matches_type(AsyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

    @parametrize
    async def test_raw_response_retrieve_ledger(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.balances.with_raw_response.retrieve_ledger()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        balance = await response.parse()
        assert_matches_type(AsyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve_ledger(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.balances.with_streaming_response.retrieve_ledger() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            balance = await response.parse()
            assert_matches_type(AsyncDefaultPageNumberPagination[BalanceLedgerEntry], balance, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["RefundStatus"]

RefundStatus: TypeAlias = Literal["succeeded", "failed", "pending", "review"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .balances import (
    BalancesResource,
    AsyncBalancesResource,
    BalancesResourceWithRawResponse,
    AsyncBalancesResourceWithRawResponse,
    BalancesResourceWithStreamingResponse,
    AsyncBalancesResourceWithStreamingResponse,
)
from .credit_entitlements import (
    CreditEntitlementsResource,
    AsyncCreditEntitlementsResource,
    CreditEntitlementsResourceWithRawResponse,
    AsyncCreditEntitlementsResourceWithRawResponse,
    CreditEntitlementsResourceWithStreamingResponse,
    AsyncCreditEntitlementsResourceWithStreamingResponse,
)

__all__ = [
    "BalancesResource",
    "AsyncBalancesResource",
    "BalancesResourceWithRawResponse",
    "AsyncBalancesResourceWithRawResponse",
    "BalancesResourceWithStreamingResponse",
    "AsyncBalancesResourceWithStreamingResponse",
    "CreditEntitlementsResource",
    "AsyncCreditEntitlementsResource",
    "CreditEntitlementsResourceWithRawResponse",
    "AsyncCreditEntitlementsResourceWithRawResponse",
    "CreditEntitlementsResourceWithStreamingResponse",
    "AsyncCreditEntitlementsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ["MeterListParams"]


class MeterListParams(TypedDict, total=False):
    archived: bool
    """List archived meters"""

    page_number: int
    """Page number default is 0"""

    page_size: int
    """Page size default is 10 max is 100"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Optional
from datetime import datetime
from typing_extensions import Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["SubscriptionRetrieveUsageHistoryParams"]


class SubscriptionRetrieveUsageHistoryParams(TypedDict, total=False):
    end_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """Filter by end date (inclusive)"""

    meter_id: Optional[str]
    """Filter by specific meter ID"""

    page_number: Optional[int]
    """Page number (default: 0)"""

    page_size: Optional[int]
    """Page size (default: 10, max: 100)"""

    start_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """Filter by start date (inclusive)"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from ..._models import BaseModel
from .ledger_entry_type import LedgerEntryType

__all__ = ["BalanceCreateLedgerEntryResponse"]


class BalanceCreateLedgerEntryResponse(BaseModel):
    """Response for creating a ledger entry"""

    id: str

    amount: str

    balance_after: str

    balance_before: str

    created_at: datetime

    credit_entitlement_id: str

    customer_id: str

    entry_type: LedgerEntryType

    is_credit: bool

    overage_after: str

    overage_before: str

    grant_id: Optional[str] = None

    reason: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["CbbProrationBehavior"]

CbbProrationBehavior: TypeAlias = Literal["prorate", "no_prorate"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .image_update_params import ImageUpdateParams as ImageUpdateParams
from .image_update_response import ImageUpdateResponse as ImageUpdateResponse
from .short_link_list_params import ShortLinkListParams as ShortLinkListParams
from .short_link_create_params import ShortLinkCreateParams as ShortLinkCreateParams
from .short_link_list_response import ShortLinkListResponse as ShortLinkListResponse
from .short_link_create_response import ShortLinkCreateResponse as ShortLinkCreateResponse



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime

from .._models import BaseModel
from .currency import Currency
from .intent_status import IntentStatus
from .dispute_status import DisputeStatus
from .payment_refund_status import PaymentRefundStatus
from .customer_limited_details import CustomerLimitedDetails

__all__ = ["PaymentListResponse"]


class PaymentListResponse(BaseModel):
    brand_id: str

    created_at: datetime

    currency: Currency

    customer: CustomerLimitedDetails

    digital_products_delivered: bool

    has_license_key: bool

    metadata: Dict[str, str]

    payment_id: str

    total_amount: int

    dispute_status: Optional[DisputeStatus] = None
    """The most recent dispute status for this payment. None if no disputes exist."""

    invoice_id: Optional[str] = None
    """Invoice ID for this payment. Uses India-specific invoice ID if available."""

    invoice_url: Optional[str] = None
    """URL to download the invoice PDF for this payment."""

    payment_method: Optional[str] = None

    payment_method_type: Optional[str] = None

    refund_status: Optional[PaymentRefundStatus] = None
    """Summary of the refund status for this payment.

    None if no succeeded refunds exist.
    """

    status: Optional[IntentStatus] = None

    subscription_id: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel
from .subscription import Subscription

__all__ = ["SubscriptionActiveWebhookEvent"]


class SubscriptionActiveWebhookEvent(BaseModel):
    business_id: str
    """The business identifier"""

    data: Subscription
    """Response struct representing subscription details"""

    timestamp: datetime
    """The timestamp of when the event occurred"""

    type: Literal["subscription.active"]
    """The event type"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .._models import BaseModel
from .currency import Currency

__all__ = ["MeterCartResponseItem"]


class MeterCartResponseItem(BaseModel):
    """Response struct representing usage-based meter cart details for a subscription"""

    currency: Currency

    free_threshold: int

    measurement_unit: str

    meter_id: str

    name: str

    description: Optional[str] = None

    price_per_unit: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["CustomFieldResponse"]


class CustomFieldResponse(BaseModel):
    """Customer's response to a custom field"""

    key: str
    """Key matching the custom field definition"""

    value: str
    """Value provided by customer"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Optional
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["EventInputParam"]


class EventInputParam(TypedDict, total=False):
    customer_id: Required[str]
    """customer_id of the customer whose usage needs to be tracked"""

    event_id: Required[str]
    """Event Id acts as an idempotency key.

    Any subsequent requests with the same event_id will be ignored
    """

    event_name: Required[str]
    """Name of the event"""

    metadata: Optional[Dict[str, Union[str, float, bool]]]
    """Custom metadata.

    Only key value pairs are accepted, objects or arrays submitted will be rejected.
    """

    timestamp: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """Custom Timestamp.

    Defaults to current timestamp in UTC. Timestamps that are older that 1 hour or
    after 5 mins, from current timestamp, will be rejected.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable
from typing_extensions import Required, TypedDict

from .conjunction import Conjunction
from .filter_operator import FilterOperator

__all__ = [
    "MeterFilterParam",
    "ClausesDirectFilterCondition",
    "ClausesNestedMeterFilter",
    "ClausesNestedMeterFilterClausesLevel1FilterCondition",
    "ClausesNestedMeterFilterClausesLevel1NestedFilter",
    "ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2FilterCondition",
    "ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2NestedFilter",
    "ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2NestedFilterClause",
]


class ClausesDirectFilterCondition(TypedDict, total=False):
    """Filter condition with key, operator, and value"""

    key: Required[str]
    """Filter key to apply"""

    operator: Required[FilterOperator]

    value: Required[Union[str, float, bool]]
    """Filter value - can be string, number, or boolean"""


class ClausesNestedMeterFilterClausesLevel1FilterCondition(TypedDict, total=False):
    """Filter condition with key, operator, and value"""

    key: Required[str]
    """Filter key to apply"""

    operator: Required[FilterOperator]

    value: Required[Union[str, float, bool]]
    """Filter value - can be string, number, or boolean"""


class ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2FilterCondition(TypedDict, total=False):
    """Filter condition with key, operator, and value"""

    key: Required[str]
    """Filter key to apply"""

    operator: Required[FilterOperator]

    value: Required[Union[str, float, bool]]
    """Filter value - can be string, number, or boolean"""


class ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2NestedFilterClause(TypedDict, total=False):
    """Filter condition with key, operator, and value"""

    key: Required[str]
    """Filter key to apply"""

    operator: Required[FilterOperator]

    value: Required[Union[str, float, bool]]
    """Filter value - can be string, number, or boolean"""


class ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2NestedFilter(TypedDict, total=False):
    """Level 3 nested filter (final nesting level)"""

    clauses: Required[Iterable[ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2NestedFilterClause]]
    """Level 3: Filter conditions only (max depth reached)"""

    conjunction: Required[Conjunction]


class ClausesNestedMeterFilterClausesLevel1NestedFilter(TypedDict, total=False):
    """Level 2 nested filter"""

    clauses: Required[
        Union[
            Iterable[ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2FilterCondition],
            Iterable[ClausesNestedMeterFilterClausesLevel1NestedFilterClausesLevel2NestedFilter],
        ]
    ]
    """Level 2: Can be conditions or nested filters (1 more level allowed)"""

    conjunction: Required[Conjunction]


class ClausesNestedMeterFilter(TypedDict, total=False):
    """Level 1 nested filter - can contain Level 2 filters"""

    clauses: Required[
        Union[
            Iterable[ClausesNestedMeterFilterClausesLevel1FilterCondition],
            Iterable[ClausesNestedMeterFilterClausesLevel1NestedFilter],
        ]
    ]
    """Level 1: Can be conditions or nested filters (2 more levels allowed)"""

    conjunction: Required[Conjunction]


class MeterFilterParam(TypedDict, total=False):
    """
    A filter structure that combines multiple conditions with logical conjunctions (AND/OR).

    Supports up to 3 levels of nesting to create complex filter expressions.
    Each filter has a conjunction (and/or) and clauses that can be either direct conditions or nested filters.
    """

    clauses: Required[Union[Iterable[ClausesDirectFilterCondition], Iterable[ClausesNestedMeterFilter]]]
    """
    Filter clauses - can be direct conditions or nested filters (up to 3 levels
    deep)
    """

    conjunction: Required[Conjunction]
    """Logical conjunction to apply between clauses (and/or)"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel
from .country_code import CountryCode
from .payment_method_types import PaymentMethodTypes

__all__ = ["CustomerRetrievePaymentMethodsResponse", "Item", "ItemCard"]


class ItemCard(BaseModel):
    card_holder_name: Optional[str] = None

    card_issuing_country: Optional[CountryCode] = None
    """ISO country code alpha2 variant"""

    card_network: Optional[str] = None

    card_type: Optional[str] = None

    expiry_month: Optional[str] = None

    expiry_year: Optional[str] = None

    last4_digits: Optional[str] = None


class Item(BaseModel):
    payment_method: Literal[
        "card",
        "card_redirect",
        "pay_later",
        "wallet",
        "bank_redirect",
        "bank_transfer",
        "crypto",
        "bank_debit",
        "reward",
        "real_time_payment",
        "upi",
        "voucher",
        "gift_card",
        "open_banking",
        "mobile_payment",
    ]

    payment_method_id: str

    card: Optional[ItemCard] = None

    last_used_at: Optional[datetime] = None

    payment_method_type: Optional[PaymentMethodTypes] = None

    recurring_enabled: Optional[bool] = None


class CustomerRetrievePaymentMethodsResponse(BaseModel):
    items: List[Item]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["BalanceListGrantsParams"]


class BalanceListGrantsParams(TypedDict, total=False):
    credit_entitlement_id: Required[str]

    page_number: int
    """Page number default is 0"""

    page_size: int
    """Page size default is 10 max is 100"""

    status: Literal["active", "expired", "depleted"]
    """Filter by grant status: active, expired, depleted"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import httpx

from ..._types import Body, Query, Headers, NotGiven, not_given
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    BinaryAPIResponse,
    AsyncBinaryAPIResponse,
    StreamedBinaryAPIResponse,
    AsyncStreamedBinaryAPIResponse,
    to_custom_raw_response_wrapper,
    to_custom_streamed_response_wrapper,
    async_to_custom_raw_response_wrapper,
    async_to_custom_streamed_response_wrapper,
)
from ..._base_client import make_request_options

__all__ = ["PaymentsResource", "AsyncPaymentsResource"]


class PaymentsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> PaymentsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return PaymentsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> PaymentsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return PaymentsResourceWithStreamingResponse(self)

    def retrieve(
        self,
        payment_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BinaryAPIResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not payment_id:
            raise ValueError(f"Expected a non-empty value for `payment_id` but received {payment_id!r}")
        extra_headers = {"Accept": "application/pdf", **(extra_headers or {})}
        return self._get(
            f"/invoices/payments/{payment_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BinaryAPIResponse,
        )

    def retrieve_refund(
        self,
        refund_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BinaryAPIResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not refund_id:
            raise ValueError(f"Expected a non-empty value for `refund_id` but received {refund_id!r}")
        extra_headers = {"Accept": "application/pdf", **(extra_headers or {})}
        return self._get(
            f"/invoices/refunds/{refund_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BinaryAPIResponse,
        )


class AsyncPaymentsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncPaymentsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return AsyncPaymentsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncPaymentsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return AsyncPaymentsResourceWithStreamingResponse(self)

    async def retrieve(
        self,
        payment_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncBinaryAPIResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not payment_id:
            raise ValueError(f"Expected a non-empty value for `payment_id` but received {payment_id!r}")
        extra_headers = {"Accept": "application/pdf", **(extra_headers or {})}
        return await self._get(
            f"/invoices/payments/{payment_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AsyncBinaryAPIResponse,
        )

    async def retrieve_refund(
        self,
        refund_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncBinaryAPIResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not refund_id:
            raise ValueError(f"Expected a non-empty value for `refund_id` but received {refund_id!r}")
        extra_headers = {"Accept": "application/pdf", **(extra_headers or {})}
        return await self._get(
            f"/invoices/refunds/{refund_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AsyncBinaryAPIResponse,
        )


class PaymentsResourceWithRawResponse:
    def __init__(self, payments: PaymentsResource) -> None:
        self._payments = payments

        self.retrieve = to_custom_raw_response_wrapper(
            payments.retrieve,
            BinaryAPIResponse,
        )
        self.retrieve_refund = to_custom_raw_response_wrapper(
            payments.retrieve_refund,
            BinaryAPIResponse,
        )


class AsyncPaymentsResourceWithRawResponse:
    def __init__(self, payments: AsyncPaymentsResource) -> None:
        self._payments = payments

        self.retrieve = async_to_custom_raw_response_wrapper(
            payments.retrieve,
            AsyncBinaryAPIResponse,
        )
        self.retrieve_refund = async_to_custom_raw_response_wrapper(
            payments.retrieve_refund,
            AsyncBinaryAPIResponse,
        )


class PaymentsResourceWithStreamingResponse:
    def __init__(self, payments: PaymentsResource) -> None:
        self._payments = payments

        self.retrieve = to_custom_streamed_response_wrapper(
            payments.retrieve,
            StreamedBinaryAPIResponse,
        )
        self.retrieve_refund = to_custom_streamed_response_wrapper(
            payments.retrieve_refund,
            StreamedBinaryAPIResponse,
        )


class AsyncPaymentsResourceWithStreamingResponse:
    def __init__(self, payments: AsyncPaymentsResource) -> None:
        self._payments = payments

        self.retrieve = async_to_custom_streamed_response_wrapper(
            payments.retrieve,
            AsyncStreamedBinaryAPIResponse,
        )
        self.retrieve_refund = async_to_custom_streamed_response_wrapper(
            payments.retrieve_refund,
            AsyncStreamedBinaryAPIResponse,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal, TypedDict

from .theme_config_param import ThemeConfigParam

__all__ = ["CheckoutSessionCustomizationParam"]


class CheckoutSessionCustomizationParam(TypedDict, total=False):
    force_language: Optional[str]
    """Force the checkout interface to render in a specific language (e.g. `en`, `es`)"""

    show_on_demand_tag: bool
    """Show on demand tag

    Default is true
    """

    show_order_details: bool
    """Show order details by default

    Default is true
    """

    theme: Literal["dark", "light", "system"]
    """Theme of the page (determines which mode - light/dark/system - to use)

    Default is `System`.
    """

    theme_config: Optional[ThemeConfigParam]
    """Optional custom theme configuration with colors for light and dark modes"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .misc import (
    MiscResource,
    AsyncMiscResource,
    MiscResourceWithRawResponse,
    AsyncMiscResourceWithRawResponse,
    MiscResourceWithStreamingResponse,
    AsyncMiscResourceWithStreamingResponse,
)
from .addons import (
    AddonsResource,
    AsyncAddonsResource,
    AddonsResourceWithRawResponse,
    AsyncAddonsResourceWithRawResponse,
    AddonsResourceWithStreamingResponse,
    AsyncAddonsResourceWithStreamingResponse,
)
from .brands import (
    BrandsResource,
    AsyncBrandsResource,
    BrandsResourceWithRawResponse,
    AsyncBrandsResourceWithRawResponse,
    BrandsResourceWithStreamingResponse,
    AsyncBrandsResourceWithStreamingResponse,
)
from .meters import (
    MetersResource,
    AsyncMetersResource,
    MetersResourceWithRawResponse,
    AsyncMetersResourceWithRawResponse,
    MetersResourceWithStreamingResponse,
    AsyncMetersResourceWithStreamingResponse,
)
from .payouts import (
    PayoutsResource,
    AsyncPayoutsResource,
    PayoutsResourceWithRawResponse,
    AsyncPayoutsResourceWithRawResponse,
    PayoutsResourceWithStreamingResponse,
    AsyncPayoutsResourceWithStreamingResponse,
)
from .refunds import (
    RefundsResource,
    AsyncRefundsResource,
    RefundsResourceWithRawResponse,
    AsyncRefundsResourceWithRawResponse,
    RefundsResourceWithStreamingResponse,
    AsyncRefundsResourceWithStreamingResponse,
)
from .balances import (
    BalancesResource,
    AsyncBalancesResource,
    BalancesResourceWithRawResponse,
    AsyncBalancesResourceWithRawResponse,
    BalancesResourceWithStreamingResponse,
    AsyncBalancesResourceWithStreamingResponse,
)
from .disputes import (
    DisputesResource,
    AsyncDisputesResource,
    DisputesResourceWithRawResponse,
    AsyncDisputesResourceWithRawResponse,
    DisputesResourceWithStreamingResponse,
    AsyncDisputesResourceWithStreamingResponse,
)
from .invoices import (
    InvoicesResource,
    AsyncInvoicesResource,
    InvoicesResourceWithRawResponse,
    AsyncInvoicesResourceWithRawResponse,
    InvoicesResourceWithStreamingResponse,
    AsyncInvoicesResourceWithStreamingResponse,
)
from .licenses import (
    LicensesResource,
    AsyncLicensesResource,
    LicensesResourceWithRawResponse,
    AsyncLicensesResourceWithRawResponse,
    LicensesResourceWithStreamingResponse,
    AsyncLicensesResourceWithStreamingResponse,
)
from .payments import (
    PaymentsResource,
    AsyncPaymentsResource,
    PaymentsResourceWithRawResponse,
    AsyncPaymentsResourceWithRawResponse,
    PaymentsResourceWithStreamingResponse,
    AsyncPaymentsResourceWithStreamingResponse,
)
from .products import (
    ProductsResource,
    AsyncProductsResource,
    ProductsResourceWithRawResponse,
    AsyncProductsResourceWithRawResponse,
    ProductsResourceWithStreamingResponse,
    AsyncProductsResourceWithStreamingResponse,
)
from .webhooks import (
    WebhooksResource,
    AsyncWebhooksResource,
    WebhooksResourceWithRawResponse,
    AsyncWebhooksResourceWithRawResponse,
    WebhooksResourceWithStreamingResponse,
    AsyncWebhooksResourceWithStreamingResponse,
)
from .customers import (
    CustomersResource,
    AsyncCustomersResource,
    CustomersResourceWithRawResponse,
    AsyncCustomersResourceWithRawResponse,
    CustomersResourceWithStreamingResponse,
    AsyncCustomersResourceWithStreamingResponse,
)
from .discounts import (
    DiscountsResource,
    AsyncDiscountsResource,
    DiscountsResourceWithRawResponse,
    AsyncDiscountsResourceWithRawResponse,
    DiscountsResourceWithStreamingResponse,
    AsyncDiscountsResourceWithStreamingResponse,
)
from .license_keys import (
    LicenseKeysResource,
    AsyncLicenseKeysResource,
    LicenseKeysResourceWithRawResponse,
    AsyncLicenseKeysResourceWithRawResponse,
    LicenseKeysResourceWithStreamingResponse,
    AsyncLicenseKeysResourceWithStreamingResponse,
)
from .usage_events import (
    UsageEventsResource,
    AsyncUsageEventsResource,
    UsageEventsResourceWithRawResponse,
    AsyncUsageEventsResourceWithRawResponse,
    UsageEventsResourceWithStreamingResponse,
    AsyncUsageEventsResourceWithStreamingResponse,
)
from .subscriptions import (
    SubscriptionsResource,
    AsyncSubscriptionsResource,
    SubscriptionsResourceWithRawResponse,
    AsyncSubscriptionsResourceWithRawResponse,
    SubscriptionsResourceWithStreamingResponse,
    AsyncSubscriptionsResourceWithStreamingResponse,
)
from .checkout_sessions import (
    CheckoutSessionsResource,
    AsyncCheckoutSessionsResource,
    CheckoutSessionsResourceWithRawResponse,
    AsyncCheckoutSessionsResourceWithRawResponse,
    CheckoutSessionsResourceWithStreamingResponse,
    AsyncCheckoutSessionsResourceWithStreamingResponse,
)
from .credit_entitlements import (
    CreditEntitlementsResource,
    AsyncCreditEntitlementsResource,
    CreditEntitlementsResourceWithRawResponse,
    AsyncCreditEntitlementsResourceWithRawResponse,
    CreditEntitlementsResourceWithStreamingResponse,
    AsyncCreditEntitlementsResourceWithStreamingResponse,
)
from .license_key_instances import (
    LicenseKeyInstancesResource,
    AsyncLicenseKeyInstancesResource,
    LicenseKeyInstancesResourceWithRawResponse,
    AsyncLicenseKeyInstancesResourceWithRawResponse,
    LicenseKeyInstancesResourceWithStreamingResponse,
    AsyncLicenseKeyInstancesResourceWithStreamingResponse,
)

__all__ = [
    "CheckoutSessionsResource",
    "AsyncCheckoutSessionsResource",
    "CheckoutSessionsResourceWithRawResponse",
    "AsyncCheckoutSessionsResourceWithRawResponse",
    "CheckoutSessionsResourceWithStreamingResponse",
    "AsyncCheckoutSessionsResourceWithStreamingResponse",
    "PaymentsResource",
    "AsyncPaymentsResource",
    "PaymentsResourceWithRawResponse",
    "AsyncPaymentsResourceWithRawResponse",
    "PaymentsResourceWithStreamingResponse",
    "AsyncPaymentsResourceWithStreamingResponse",
    "SubscriptionsResource",
    "AsyncSubscriptionsResource",
    "SubscriptionsResourceWithRawResponse",
    "AsyncSubscriptionsResourceWithRawResponse",
    "SubscriptionsResourceWithStreamingResponse",
    "AsyncSubscriptionsResourceWithStreamingResponse",
    "InvoicesResource",
    "AsyncInvoicesResource",
    "InvoicesResourceWithRawResponse",
    "AsyncInvoicesResourceWithRawResponse",
    "InvoicesResourceWithStreamingResponse",
    "AsyncInvoicesResourceWithStreamingResponse",
    "LicensesResource",
    "AsyncLicensesResource",
    "LicensesResourceWithRawResponse",
    "AsyncLicensesResourceWithRawResponse",
    "LicensesResourceWithStreamingResponse",
    "AsyncLicensesResourceWithStreamingResponse",
    "LicenseKeysResource",
    "AsyncLicenseKeysResource",
    "LicenseKeysResourceWithRawResponse",
    "AsyncLicenseKeysResourceWithRawResponse",
    "LicenseKeysResourceWithStreamingResponse",
    "AsyncLicenseKeysResourceWithStreamingResponse",
    "LicenseKeyInstancesResource",
    "AsyncLicenseKeyInstancesResource",
    "LicenseKeyInstancesResourceWithRawResponse",
    "AsyncLicenseKeyInstancesResourceWithRawResponse",
    "LicenseKeyInstancesResourceWithStreamingResponse",
    "AsyncLicenseKeyInstancesResourceWithStreamingResponse",
    "CustomersResource",
    "AsyncCustomersResource",
    "CustomersResourceWithRawResponse",
    "AsyncCustomersResourceWithRawResponse",
    "CustomersResourceWithStreamingResponse",
    "AsyncCustomersResourceWithStreamingResponse",
    "RefundsResource",
    "AsyncRefundsResource",
    "RefundsResourceWithRawResponse",
    "AsyncRefundsResourceWithRawResponse",
    "RefundsResourceWithStreamingResponse",
    "AsyncRefundsResourceWithStreamingResponse",
    "DisputesResource",
    "AsyncDisputesResource",
    "DisputesResourceWithRawResponse",
    "AsyncDisputesResourceWithRawResponse",
    "DisputesResourceWithStreamingResponse",
    "AsyncDisputesResourceWithStreamingResponse",
    "PayoutsResource",
    "AsyncPayoutsResource",
    "PayoutsResourceWithRawResponse",
    "AsyncPayoutsResourceWithRawResponse",
    "PayoutsResourceWithStreamingResponse",
    "AsyncPayoutsResourceWithStreamingResponse",
    "ProductsResource",
    "AsyncProductsResource",
    "ProductsResourceWithRawResponse",
    "AsyncProductsResourceWithRawResponse",
    "ProductsResourceWithStreamingResponse",
    "AsyncProductsResourceWithStreamingResponse",
    "MiscResource",
    "AsyncMiscResource",
    "MiscResourceWithRawResponse",
    "AsyncMiscResourceWithRawResponse",
    "MiscResourceWithStreamingResponse",
    "AsyncMiscResourceWithStreamingResponse",
    "DiscountsResource",
    "AsyncDiscountsResource",
    "DiscountsResourceWithRawResponse",
    "AsyncDiscountsResourceWithRawResponse",
    "DiscountsResourceWithStreamingResponse",
    "AsyncDiscountsResourceWithStreamingResponse",
    "AddonsResource",
    "AsyncAddonsResource",
    "AddonsResourceWithRawResponse",
    "AsyncAddonsResourceWithRawResponse",
    "AddonsResourceWithStreamingResponse",
    "AsyncAddonsResourceWithStreamingResponse",
    "BrandsResource",
    "AsyncBrandsResource",
    "BrandsResourceWithRawResponse",
    "AsyncBrandsResourceWithRawResponse",
    "BrandsResourceWithStreamingResponse",
    "AsyncBrandsResourceWithStreamingResponse",
    "WebhooksResource",
    "AsyncWebhooksResource",
    "WebhooksResourceWithRawResponse",
    "AsyncWebhooksResourceWithRawResponse",
    "WebhooksResourceWithStreamingResponse",
    "AsyncWebhooksResourceWithStreamingResponse",
    "UsageEventsResource",
    "AsyncUsageEventsResource",
    "UsageEventsResourceWithRawResponse",
    "AsyncUsageEventsResourceWithRawResponse",
    "UsageEventsResourceWithStreamingResponse",
    "AsyncUsageEventsResourceWithStreamingResponse",
    "MetersResource",
    "AsyncMetersResource",
    "MetersResourceWithRawResponse",
    "AsyncMetersResourceWithRawResponse",
    "MetersResourceWithStreamingResponse",
    "AsyncMetersResourceWithStreamingResponse",
    "BalancesResource",
    "AsyncBalancesResource",
    "BalancesResourceWithRawResponse",
    "AsyncBalancesResourceWithRawResponse",
    "BalancesResourceWithStreamingResponse",
    "AsyncBalancesResourceWithStreamingResponse",
    "CreditEntitlementsResource",
    "AsyncCreditEntitlementsResource",
    "CreditEntitlementsResourceWithRawResponse",
    "AsyncCreditEntitlementsResourceWithRawResponse",
    "CreditEntitlementsResourceWithStreamingResponse",
    "AsyncCreditEntitlementsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Optional
from datetime import datetime

import httpx

from ...types import customer_list_params, customer_create_params, customer_update_params
from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination
from ..._base_client import AsyncPaginator, make_request_options
from .customer_portal import (
    CustomerPortalResource,
    AsyncCustomerPortalResource,
    CustomerPortalResourceWithRawResponse,
    AsyncCustomerPortalResourceWithRawResponse,
    CustomerPortalResourceWithStreamingResponse,
    AsyncCustomerPortalResourceWithStreamingResponse,
)
from .wallets.wallets import (
    WalletsResource,
    AsyncWalletsResource,
    WalletsResourceWithRawResponse,
    AsyncWalletsResourceWithRawResponse,
    WalletsResourceWithStreamingResponse,
    AsyncWalletsResourceWithStreamingResponse,
)
from ...types.customer import Customer
from ...types.customer_list_credit_entitlements_response import CustomerListCreditEntitlementsResponse
from ...types.customer_retrieve_payment_methods_response import CustomerRetrievePaymentMethodsResponse

__all__ = ["CustomersResource", "AsyncCustomersResource"]


class CustomersResource(SyncAPIResource):
    @cached_property
    def customer_portal(self) -> CustomerPortalResource:
        return CustomerPortalResource(self._client)

    @cached_property
    def wallets(self) -> WalletsResource:
        return WalletsResource(self._client)

    @cached_property
    def with_raw_response(self) -> CustomersResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return CustomersResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> CustomersResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return CustomersResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        email: str,
        name: str,
        metadata: Dict[str, str] | Omit = omit,
        phone_number: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Customer:
        """
        Args:
          metadata: Additional metadata for the customer

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/customers",
            body=maybe_transform(
                {
                    "email": email,
                    "name": name,
                    "metadata": metadata,
                    "phone_number": phone_number,
                },
                customer_create_params.CustomerCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Customer,
        )

    def retrieve(
        self,
        customer_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Customer:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get(
            f"/customers/{customer_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Customer,
        )

    def update(
        self,
        customer_id: str,
        *,
        email: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        name: Optional[str] | Omit = omit,
        phone_number: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Customer:
        """
        Args:
          metadata: Additional metadata for the customer

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._patch(
            f"/customers/{customer_id}",
            body=maybe_transform(
                {
                    "email": email,
                    "metadata": metadata,
                    "name": name,
                    "phone_number": phone_number,
                },
                customer_update_params.CustomerUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Customer,
        )

    def list(
        self,
        *,
        created_at_gte: Union[str, datetime] | Omit = omit,
        created_at_lte: Union[str, datetime] | Omit = omit,
        email: str | Omit = omit,
        name: str | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[Customer]:
        """
        Args:
          created_at_gte: Filter customers created on or after this timestamp

          created_at_lte: Filter customers created on or before this timestamp

          email: Filter by customer email

          name: Filter by customer name (partial match, case-insensitive)

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/customers",
            page=SyncDefaultPageNumberPagination[Customer],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "created_at_gte": created_at_gte,
                        "created_at_lte": created_at_lte,
                        "email": email,
                        "name": name,
                        "page_number": page_number,
                        "page_size": page_size,
                    },
                    customer_list_params.CustomerListParams,
                ),
            ),
            model=Customer,
        )

    def list_credit_entitlements(
        self,
        customer_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerListCreditEntitlementsResponse:
        """
        List all credit entitlements for a customer with their current balances

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get(
            f"/customers/{customer_id}/credit-entitlements",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerListCreditEntitlementsResponse,
        )

    def retrieve_payment_methods(
        self,
        customer_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerRetrievePaymentMethodsResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get(
            f"/customers/{customer_id}/payment-methods",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerRetrievePaymentMethodsResponse,
        )


class AsyncCustomersResource(AsyncAPIResource):
    @cached_property
    def customer_portal(self) -> AsyncCustomerPortalResource:
        return AsyncCustomerPortalResource(self._client)

    @cached_property
    def wallets(self) -> AsyncWalletsResource:
        return AsyncWalletsResource(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncCustomersResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return AsyncCustomersResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncCustomersResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return AsyncCustomersResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        email: str,
        name: str,
        metadata: Dict[str, str] | Omit = omit,
        phone_number: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Customer:
        """
        Args:
          metadata: Additional metadata for the customer

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/customers",
            body=await async_maybe_transform(
                {
                    "email": email,
                    "name": name,
                    "metadata": metadata,
                    "phone_number": phone_number,
                },
                customer_create_params.CustomerCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Customer,
        )

    async def retrieve(
        self,
        customer_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Customer:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._get(
            f"/customers/{customer_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Customer,
        )

    async def update(
        self,
        customer_id: str,
        *,
        email: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        name: Optional[str] | Omit = omit,
        phone_number: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Customer:
        """
        Args:
          metadata: Additional metadata for the customer

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._patch(
            f"/customers/{customer_id}",
            body=await async_maybe_transform(
                {
                    "email": email,
                    "metadata": metadata,
                    "name": name,
                    "phone_number": phone_number,
                },
                customer_update_params.CustomerUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Customer,
        )

    def list(
        self,
        *,
        created_at_gte: Union[str, datetime] | Omit = omit,
        created_at_lte: Union[str, datetime] | Omit = omit,
        email: str | Omit = omit,
        name: str | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[Customer, AsyncDefaultPageNumberPagination[Customer]]:
        """
        Args:
          created_at_gte: Filter customers created on or after this timestamp

          created_at_lte: Filter customers created on or before this timestamp

          email: Filter by customer email

          name: Filter by customer name (partial match, case-insensitive)

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/customers",
            page=AsyncDefaultPageNumberPagination[Customer],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "created_at_gte": created_at_gte,
                        "created_at_lte": created_at_lte,
                        "email": email,
                        "name": name,
                        "page_number": page_number,
                        "page_size": page_size,
                    },
                    customer_list_params.CustomerListParams,
                ),
            ),
            model=Customer,
        )

    async def list_credit_entitlements(
        self,
        customer_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerListCreditEntitlementsResponse:
        """
        List all credit entitlements for a customer with their current balances

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._get(
            f"/customers/{customer_id}/credit-entitlements",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerListCreditEntitlementsResponse,
        )

    async def retrieve_payment_methods(
        self,
        customer_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerRetrievePaymentMethodsResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._get(
            f"/customers/{customer_id}/payment-methods",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerRetrievePaymentMethodsResponse,
        )


class CustomersResourceWithRawResponse:
    def __init__(self, customers: CustomersResource) -> None:
        self._customers = customers

        self.create = to_raw_response_wrapper(
            customers.create,
        )
        self.retrieve = to_raw_response_wrapper(
            customers.retrieve,
        )
        self.update = to_raw_response_wrapper(
            customers.update,
        )
        self.list = to_raw_response_wrapper(
            customers.list,
        )
        self.list_credit_entitlements = to_raw_response_wrapper(
            customers.list_credit_entitlements,
        )
        self.retrieve_payment_methods = to_raw_response_wrapper(
            customers.retrieve_payment_methods,
        )

    @cached_property
    def customer_portal(self) -> CustomerPortalResourceWithRawResponse:
        return CustomerPortalResourceWithRawResponse(self._customers.customer_portal)

    @cached_property
    def wallets(self) -> WalletsResourceWithRawResponse:
        return WalletsResourceWithRawResponse(self._customers.wallets)


class AsyncCustomersResourceWithRawResponse:
    def __init__(self, customers: AsyncCustomersResource) -> None:
        self._customers = customers

        self.create = async_to_raw_response_wrapper(
            customers.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            customers.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            customers.update,
        )
        self.list = async_to_raw_response_wrapper(
            customers.list,
        )
        self.list_credit_entitlements = async_to_raw_response_wrapper(
            customers.list_credit_entitlements,
        )
        self.retrieve_payment_methods = async_to_raw_response_wrapper(
            customers.retrieve_payment_methods,
        )

    @cached_property
    def customer_portal(self) -> AsyncCustomerPortalResourceWithRawResponse:
        return AsyncCustomerPortalResourceWithRawResponse(self._customers.customer_portal)

    @cached_property
    def wallets(self) -> AsyncWalletsResourceWithRawResponse:
        return AsyncWalletsResourceWithRawResponse(self._customers.wallets)


class CustomersResourceWithStreamingResponse:
    def __init__(self, customers: CustomersResource) -> None:
        self._customers = customers

        self.create = to_streamed_response_wrapper(
            customers.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            customers.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            customers.update,
        )
        self.list = to_streamed_response_wrapper(
            customers.list,
        )
        self.list_credit_entitlements = to_streamed_response_wrapper(
            customers.list_credit_entitlements,
        )
        self.retrieve_payment_methods = to_streamed_response_wrapper(
            customers.retrieve_payment_methods,
        )

    @cached_property
    def customer_portal(self) -> CustomerPortalResourceWithStreamingResponse:
        return CustomerPortalResourceWithStreamingResponse(self._customers.customer_portal)

    @cached_property
    def wallets(self) -> WalletsResourceWithStreamingResponse:
        return WalletsResourceWithStreamingResponse(self._customers.wallets)


class AsyncCustomersResourceWithStreamingResponse:
    def __init__(self, customers: AsyncCustomersResource) -> None:
        self._customers = customers

        self.create = async_to_streamed_response_wrapper(
            customers.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            customers.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            customers.update,
        )
        self.list = async_to_streamed_response_wrapper(
            customers.list,
        )
        self.list_credit_entitlements = async_to_streamed_response_wrapper(
            customers.list_credit_entitlements,
        )
        self.retrieve_payment_methods = async_to_streamed_response_wrapper(
            customers.retrieve_payment_methods,
        )

    @cached_property
    def customer_portal(self) -> AsyncCustomerPortalResourceWithStreamingResponse:
        return AsyncCustomerPortalResourceWithStreamingResponse(self._customers.customer_portal)

    @cached_property
    def wallets(self) -> AsyncWalletsResourceWithStreamingResponse:
        return AsyncWalletsResourceWithStreamingResponse(self._customers.wallets)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List
from datetime import datetime

from .._models import BaseModel
from .currency import Currency

__all__ = ["SubscriptionRetrieveUsageHistoryResponse", "Meter"]


class Meter(BaseModel):
    id: str
    """Meter identifier"""

    chargeable_units: str
    """Chargeable units (after free threshold) as string for precision"""

    consumed_units: str
    """Total units consumed as string for precision"""

    currency: Currency
    """Currency for the price per unit"""

    free_threshold: int
    """Free threshold units for this meter"""

    name: str
    """Meter name"""

    price_per_unit: str
    """Price per unit in string format for precision"""

    total_price: int
    """Total price charged for this meter in smallest currency unit (cents)"""


class SubscriptionRetrieveUsageHistoryResponse(BaseModel):
    end_date: datetime
    """End date of the billing period"""

    meters: List[Meter]
    """List of meters and their usage for this billing period"""

    start_date: datetime
    """Start date of the billing period"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from .._models import BaseModel

__all__ = ["CustomerListCreditEntitlementsResponse", "Item"]


class Item(BaseModel):
    """A credit entitlement with the customer's current balance"""

    balance: str
    """Customer's current remaining credit balance"""

    credit_entitlement_id: str
    """Credit entitlement ID"""

    name: str
    """Name of the credit entitlement"""

    overage: str
    """Customer's current overage balance"""

    unit: str
    """Unit label (e.g. "API Calls", "Tokens")"""

    description: Optional[str] = None
    """Description of the credit entitlement"""


class CustomerListCreditEntitlementsResponse(BaseModel):
    items: List[Item]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from ..._models import BaseModel

__all__ = ["ShortLinkCreateResponse"]


class ShortLinkCreateResponse(BaseModel):
    full_url: str
    """Full URL."""

    short_url: str
    """Short URL."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional

from .._models import BaseModel

__all__ = ["WebhookDetails"]


class WebhookDetails(BaseModel):
    id: str
    """The webhook's ID."""

    created_at: str
    """Created at timestamp"""

    description: str
    """An example webhook name."""

    metadata: Dict[str, str]
    """Metadata of the webhook"""

    updated_at: str
    """Updated at timestamp"""

    url: str
    """Url endpoint of the webhook"""

    disabled: Optional[bool] = None
    """Status of the webhook.

    If true, events are not sent
    """

    filter_types: Optional[List[str]] = None
    """Filter events to the webhook.

    Webhook event will only be sent for events in the list.
    """

    rate_limit: Optional[int] = None
    """Configured rate limit"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal, TypedDict

from .theme_mode_config_param import ThemeModeConfigParam

__all__ = ["ThemeConfigParam"]


class ThemeConfigParam(TypedDict, total=False):
    """Custom theme configuration with colors for light and dark modes."""

    dark: Optional[ThemeModeConfigParam]
    """Dark mode color configuration"""

    font_primary_url: Optional[str]
    """URL for the primary font. Must be a valid https:// URL."""

    font_secondary_url: Optional[str]
    """URL for the secondary font. Must be a valid https:// URL."""

    font_size: Optional[Literal["xs", "sm", "md", "lg", "xl", "2xl"]]
    """Font size for the checkout UI"""

    font_weight: Optional[Literal["normal", "medium", "bold", "extraBold"]]
    """Font weight for the checkout UI"""

    light: Optional[ThemeModeConfigParam]
    """Light mode color configuration"""

    pay_button_text: Optional[str]
    """Custom text for the pay button (e.g., "Complete Purchase", "Subscribe Now").

    Max 100 characters.
    """

    radius: Optional[str]
    """Border radius for UI elements.

    Must be a number followed by px, rem, or em (e.g., "4px", "0.5rem", "1em")
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from .._models import BaseModel

__all__ = ["SubscriptionUpdatePaymentMethodResponse"]


class SubscriptionUpdatePaymentMethodResponse(BaseModel):
    client_secret: Optional[str] = None

    expires_on: Optional[datetime] = None

    payment_id: Optional[str] = None

    payment_link: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel
from .credit_entitlements.credit_ledger_entry import CreditLedgerEntry

__all__ = ["CreditDeductedWebhookEvent"]


class CreditDeductedWebhookEvent(BaseModel):
    business_id: str
    """The business identifier"""

    data: CreditLedgerEntry
    """Response for a ledger entry"""

    timestamp: datetime
    """The timestamp of when the event occurred"""

    type: Literal["credit.deducted"]
    """The event type"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ["ShortLinkListParams"]


class ShortLinkListParams(TypedDict, total=False):
    page_number: int
    """Page number default is 0"""

    page_size: int
    """Page size default is 10 max is 100"""

    product_id: str
    """Filter by product ID"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Optional
from typing_extensions import Literal, Required, TypeAlias, TypedDict

__all__ = ["SubscriptionUpdatePaymentMethodParams", "New", "Existing"]


class New(TypedDict, total=False):
    type: Required[Literal["new"]]

    return_url: Optional[str]


class Existing(TypedDict, total=False):
    payment_method_id: Required[str]

    type: Required[Literal["existing"]]


SubscriptionUpdatePaymentMethodParams: TypeAlias = Union[New, Existing]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from typing_extensions import Literal, TypeAlias

from .._models import BaseModel
from .currency import Currency
from .subscription import Subscription
from .tax_category import TaxCategory

__all__ = [
    "SubscriptionPreviewChangePlanResponse",
    "ImmediateCharge",
    "ImmediateChargeLineItem",
    "ImmediateChargeLineItemSubscription",
    "ImmediateChargeLineItemAddon",
    "ImmediateChargeLineItemMeter",
    "ImmediateChargeSummary",
]


class ImmediateChargeLineItemSubscription(BaseModel):
    id: str

    currency: Currency

    product_id: str

    proration_factor: float

    quantity: int

    tax_inclusive: bool

    type: Literal["subscription"]

    unit_price: int

    description: Optional[str] = None

    name: Optional[str] = None

    tax: Optional[int] = None

    tax_rate: Optional[float] = None


class ImmediateChargeLineItemAddon(BaseModel):
    id: str

    currency: Currency

    name: str

    proration_factor: float

    quantity: int

    tax_category: TaxCategory
    """
    Represents the different categories of taxation applicable to various products
    and services.
    """

    tax_inclusive: bool

    tax_rate: float

    type: Literal["addon"]

    unit_price: int

    description: Optional[str] = None

    tax: Optional[int] = None


class ImmediateChargeLineItemMeter(BaseModel):
    id: str

    chargeable_units: str

    currency: Currency

    free_threshold: int

    name: str

    price_per_unit: str

    subtotal: int

    tax_inclusive: bool

    tax_rate: float

    type: Literal["meter"]

    units_consumed: str

    description: Optional[str] = None

    tax: Optional[int] = None


ImmediateChargeLineItem: TypeAlias = Union[
    ImmediateChargeLineItemSubscription, ImmediateChargeLineItemAddon, ImmediateChargeLineItemMeter
]


class ImmediateChargeSummary(BaseModel):
    currency: Currency

    customer_credits: int

    settlement_amount: int

    settlement_currency: Currency

    total_amount: int

    settlement_tax: Optional[int] = None

    tax: Optional[int] = None


class ImmediateCharge(BaseModel):
    line_items: List[ImmediateChargeLineItem]

    summary: ImmediateChargeSummary


class SubscriptionPreviewChangePlanResponse(BaseModel):
    immediate_charge: ImmediateCharge

    new_plan: Subscription
    """Response struct representing subscription details"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ["CheckoutSessionFlagsParam"]


class CheckoutSessionFlagsParam(TypedDict, total=False):
    allow_currency_selection: bool
    """if customer is allowed to change currency, set it to true

    Default is true
    """

    allow_customer_editing_city: bool

    allow_customer_editing_country: bool

    allow_customer_editing_email: bool

    allow_customer_editing_name: bool

    allow_customer_editing_state: bool

    allow_customer_editing_street: bool

    allow_customer_editing_tax_id: bool

    allow_customer_editing_zipcode: bool

    allow_discount_code: bool
    """If the customer is allowed to apply discount code, set it to true.

    Default is true
    """

    allow_phone_number_collection: bool
    """If phone number is collected from customer, set it to rue

    Default is true
    """

    allow_tax_id: bool
    """If the customer is allowed to add tax id, set it to true

    Default is true
    """

    always_create_new_customer: bool
    """
    Set to true if a new customer object should be created. By default email is used
    to find an existing customer to attach the session to

    Default is false
    """

    redirect_immediately: bool
    """If true, redirects the customer immediately after payment completion

    Default is false
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import TypedDict

__all__ = ["ThemeModeConfigParam"]


class ThemeModeConfigParam(TypedDict, total=False):
    """Color configuration for a single theme mode (light or dark).

    All color fields accept standard CSS color formats:
    - Hex: `#fff`, `#ffffff`, `#ffffffff` (with or without # prefix)
    - RGB/RGBA: `rgb(255, 255, 255)`, `rgba(255, 255, 255, 0.5)`
    - HSL/HSLA: `hsl(120, 100%, 50%)`, `hsla(120, 100%, 50%, 0.5)`
    - Named colors: `red`, `blue`, `transparent`, etc.
    - Advanced: `hwb()`, `lab()`, `lch()`, `oklab()`, `oklch()`, `color()`
    """

    bg_primary: Optional[str]
    """Background primary color

    Examples: `"#ffffff"`, `"rgb(255, 255, 255)"`, `"white"`
    """

    bg_secondary: Optional[str]
    """Background secondary color"""

    border_primary: Optional[str]
    """Border primary color"""

    border_secondary: Optional[str]
    """Border secondary color"""

    button_primary: Optional[str]
    """Primary button background color"""

    button_primary_hover: Optional[str]
    """Primary button hover color"""

    button_secondary: Optional[str]
    """Secondary button background color"""

    button_secondary_hover: Optional[str]
    """Secondary button hover color"""

    button_text_primary: Optional[str]
    """Primary button text color"""

    button_text_secondary: Optional[str]
    """Secondary button text color"""

    input_focus_border: Optional[str]
    """Input focus border color"""

    text_error: Optional[str]
    """Text error color"""

    text_placeholder: Optional[str]
    """Text placeholder color"""

    text_primary: Optional[str]
    """Text primary color"""

    text_secondary: Optional[str]
    """Text secondary color"""

    text_success: Optional[str]
    """Text success color"""
